from ta.trend import SMAIndicator

from FinMind.BackTestSystem.BaseClass import Strategy


class Bias(Strategy):
    """
    url:
    "https://bc8800.pixnet.net/blog/post/ \
        328645973-%E7%A8%8B%E5%BC%8F%E4%BA%A4 \
        %E6%98%93%E4%B9%8B%E7%A0%94%E7%A9%B6% \
        E7%AD%86%E8%A8%98---%E4%B9%96%E9%9B%A2%E7%8E%87"

    summary:
    乖離率策略是觀察股價偏離移動平均線(MA線)的程度來決定是否進場
        負乖離表示股價 低 於過去一段時間平均價，意味著股價相對過去 低 ，所以選擇進場
        正乖離表示股價 高 於過去一段時間平均價，意味著股價相對過去 高 ，所以選擇出場
    """

    ma_days = 24
    bias_lower = -7
    bias_upper = 8

    def init(self, base_data):
        base_data = base_data.sort_values("date")

        base_data[f"ma{self.ma_days}"] = SMAIndicator(
            base_data["close"], self.ma_days
        ).sma_indicator()
        base_data["bias"] = (
            (base_data["close"] - base_data[f"ma{self.ma_days}"])
            / base_data[f"ma{self.ma_days}"]
        ) * 100
        base_data = base_data.dropna()
        base_data.index = range(len(base_data))

        base_data["signal"] = base_data["bias"].map(
            lambda x: 1
            if x < self.bias_lower
            else (-1 if x > self.bias_upper else 0)
        )

        base_data["signal"] = base_data["signal"].fillna(0)
        return base_data



from FinMind.Data.Load import FinData


def get_asset_underlying_type(stock_id: str) -> str:
    TaiwanStockInfo = FinData("TaiwanStockInfo", date="")
    underlying_type = TaiwanStockInfo[TaiwanStockInfo["stock_id"] == stock_id][
        "industry_category"
    ].values[0]
    return underlying_type


def get_underlying_trading_tax(underlying_type: str) -> float:
    mapping = {"ETF": 0.001, "上櫃指數股票型基金(ETF)": 0.001}
    return mapping.get(underlying_type, 0.003)



from FinMind.BackTestSystem.BaseClass import Strategy


class ContinueHolding(Strategy):
    buy_freq_day = 30

    def init(self, base_data):
        base_data["signal"] = (base_data.index % self.buy_freq_day == 0).astype(
            int
        )
        return base_data



import pandas as pd
from FinMind.Crawler import RawMaterialFuturesPrices
from FinMind.Crawler import GovernmentBonds

# -------------------------------------------------------------------
RMFP = RawMaterialFuturesPrices.Crawler()
# get futures list
loop_list = RMFP.create_loop_list()

# get one futures data
data = RMFP.crawler(loop_list[0])

# or get all futures data
data = pd.DataFrame()
for loop in loop_list:
    print(loop)
    value = RMFP.crawler(loop)
    data = data.append(value)

# -------------------------------------------------------------------
GB = GovernmentBonds.Crawler()
# get futures list
loop_list = GB.create_loop_list()

# get one futures data
data = GB.crawler(loop_list[0])

# or get all futures data
data = pd.DataFrame()
for loop in loop_list:
    print(loop)
    value = GB.crawler(loop)
    data = data.append(value)



import typing
import warnings

import numpy as np
import pandas as pd
from FinMind.BackTestSystem.utils import (
    get_asset_underlying_type,
    get_underlying_trading_tax,
)
from FinMind.Data.Load import FinData


class Trader:
    def __init__(
        self,
        stock_id: str,
        trader_fund: int,
        hold_volume: float,
        hold_cost: float,
        fee: float,
        tax: float,
        board_lot: int = 1000,
    ):
        self.stock_id = stock_id
        self.trader_fund = trader_fund
        self.hold_volume = hold_volume
        self.hold_cost = hold_cost
        self.fee = fee
        self.tax = tax
        self.trade_price = None
        self.board_lot = board_lot
        self.UnrealizedProfit = 0
        self.RealizedProfit = 0
        self.EverytimeProfit = 0
        # TODO: EverytimeTraderFund

    def buy(self, trade_price: float, trade_lots: float):
        self.trade_price = trade_price
        if (
            self.__confirm_trade_lots(trade_lots, trade_price, self.trader_fund)
            > 0
        ):
            trade_volume = trade_lots * self.board_lot
            buy_fee = max(20, self.trade_price * trade_volume * self.fee)
            buy_price = self.trade_price * trade_volume
            buy_total_price = buy_price + buy_fee
            self.trader_fund = self.trader_fund - buy_total_price
            origin_hold_cost = self.hold_volume * self.hold_cost
            self.hold_volume = self.hold_volume + trade_volume
            self.hold_cost = (
                origin_hold_cost + buy_total_price
            ) / self.hold_volume

        self.__compute_realtime_status()

    def sell(self, trade_price: float, trade_lots: float):
        self.trade_price = trade_price
        if (
            self.__confirm_trade_lots(trade_lots, trade_price, self.trader_fund)
            > 0
        ):
            trade_volume = trade_lots * self.board_lot
            sell_fee = max(20, trade_price * trade_volume * self.fee)
            sell_tax = trade_price * trade_volume * self.tax
            sell_price = trade_price * trade_volume
            sell_total_price = sell_price - sell_tax - sell_fee
            self.trader_fund = self.trader_fund + sell_total_price
            self.RealizedProfit = self.RealizedProfit + round(
                sell_total_price - (self.hold_cost * trade_volume), 2,
            )
            self.hold_volume = self.hold_volume - trade_volume

        self.__compute_realtime_status()

    def noaction(self, trade_price: float):
        self.trade_price = trade_price
        self.__compute_realtime_status()

    def __compute_realtime_status(self):
        sell_fee = max(20, self.trade_price * self.hold_volume * self.fee)
        sell_fee = sell_fee if self.hold_volume > 0 else 0
        sell_tax = self.trade_price * self.hold_volume * self.tax
        sell_price = self.trade_price * self.hold_volume
        capital_gains = sell_price - self.hold_cost * self.hold_volume
        self.UnrealizedProfit = capital_gains - sell_fee - sell_tax
        self.EverytimeProfit = self.UnrealizedProfit + self.RealizedProfit

    def __have_enough_money(
        self, trader_fund: int, trade_price: float, trade_volume: int
    ) -> bool:
        if trader_fund < (trade_price * trade_volume):
            return False
        else:
            return True

    def __have_enough_volume(
        self, hold_volume: float, trade_volume: int
    ) -> bool:
        if hold_volume < trade_volume:
            return False
        else:
            return True

    def __confirm_trade_lots(
        self, trade_lots: int, trade_price: float, trader_fund: int
    ):
        """
        do not have enough money --> not buy
        do not have enough lots --> not sell

        # TODO: in the future can expand
        if only have 4 los money, but buing 5 lots
            --> not buy, since money not enough, as the same as sell
        """
        final_trade_lots = 0
        trade_volume = trade_lots * self.board_lot
        if trade_lots > 0:
            if self.__have_enough_money(trader_fund, trade_price, trade_volume):
                final_trade_lots = trade_lots
            else:
                final_trade_lots = 0
        elif trade_lots < 0:
            hold_volume = self.hold_volume
            if self.__have_enough_volume(hold_volume, trade_volume):
                final_trade_lots = trade_lots
            else:
                final_trade_lots = 0
        return final_trade_lots


class Strategy:
    def __init__(self, trader):
        self.trader = trader

    def trade(self, signal: float, trade_price: float):
        if signal > 0:
            self.buy(trade_price=trade_price, trade_lots=signal)
        elif signal < 0:
            self.sell(trade_price=trade_price, trade_lots=signal)
        else:
            self.noaction(trade_price=trade_price)

    def buy(self, trade_price: float, trade_lots: float):
        self.trader.buy(trade_price, trade_lots)

    def sell(self, trade_price: float, trade_lots: float):
        self.trader.sell(trade_price, trade_lots)

    def noaction(self, trade_price: float):
        self.trader.noaction(trade_price)


class BackTest:
    def __init__(
        self,
        strategy: typing.Type[Strategy],
        user_id: str = "",
        password: str = "",
        stock_id: str = "",
        start_date: str = "",
        end_date: str = "",
        trader_fund: float = 0,
        fee: float = 0.001425,
    ):
        if not (isinstance(strategy, type) and issubclass(strategy, Strategy)):
            raise TypeError("`strategy` must be a Strategy sub-type")

        self.stock_id = stock_id
        self.start_date = start_date
        self.end_date = end_date
        self.trader_fund = trader_fund
        self.fee = fee
        underlying_type = get_asset_underlying_type(stock_id)
        self.tax = get_underlying_trading_tax(underlying_type)
        self.trader = Trader(
            stock_id=stock_id,
            hold_volume=0,
            hold_cost=0,
            trader_fund=trader_fund,
            fee=self.fee,
            tax=self.tax,
        )
        self.user_id = user_id
        self.password = password
        self.strategy = strategy
        self._results = pd.DataFrame()
        self._final_stats = pd.Series()

    def __init_base_data(self) -> pd.DataFrame:
        # FIXME: some stock_id do not have div
        self.base_data = FinData(
            dataset="TaiwanStockPrice",
            select=self.stock_id,
            date=self.start_date,
            end_date=self.end_date,
            user_id=self.user_id,
            password=self.password,
        )
        StockDividend = FinData(
            dataset="StockDividend",
            select=self.stock_id,
            date=self.start_date,
            end_date=self.end_date,
            user_id=self.user_id,
            password=self.password,
        )
        if not StockDividend.empty:
            cash_div = StockDividend[
                [
                    "stock_id",
                    "CashExDividendTradingDate",
                    "CashEarningsDistribution",
                ]
            ].rename(columns={"CashExDividendTradingDate": "date"})
            stock_div = StockDividend[
                [
                    "stock_id",
                    "StockExDividendTradingDate",
                    "StockEarningsDistribution",
                ]
            ].rename(columns={"StockExDividendTradingDate": "date"})
            self.base_data = pd.merge(
                self.base_data,
                cash_div,
                left_on=["stock_id", "date"],
                right_on=["stock_id", "date"],
                how="left",
            ).fillna(0)
            self.base_data = pd.merge(
                self.base_data,
                stock_div,
                left_on=["stock_id", "date"],
                right_on=["stock_id", "date"],
                how="left",
            ).fillna(0)
        else:
            self.base_data["StockEarningsDistribution"] = 0
            self.base_data["CashEarningsDistribution"] = 0

    def simulate(self):
        self.__init_base_data()
        trader = self.trader
        strategy = self.strategy(trader)
        self.base_data = strategy.init(base_data=self.base_data)
        assert (
            "signal" in self.base_data.columns
        ), "Must be create signal columns in base_data"
        if not self.base_data.index.is_monotonic_increasing:
            warnings.warn(
                "Data index is not sorted in ascending order. Sorting.",
                stacklevel=2,
            )
            self.base_data = self.base_data.sort_index()
        for i in range(1, len(self.base_data)):
            # use last date to decide buy or sell or nothing
            last_date_index = i - 1
            signal = self.base_data.loc[last_date_index, "signal"]
            trade_price = self.base_data.loc[i, "open"]
            strategy.trade(signal, trade_price)
            cash_div = self.base_data.loc[i, "CashEarningsDistribution"]
            stock_div = self.base_data.loc[i, "StockEarningsDistribution"]
            self.__compute_div_income(strategy.trader, cash_div, stock_div)
            dic_value = strategy.trader.__dict__
            dic_value["date"] = self.base_data.loc[i, "date"]
            self._results = self._results.append(dic_value, ignore_index=True)

        self.__compute_final_stats()

    def __compute_div_income(self, trader, cash_div: float, stock_div: float):
        gain_stock_div = stock_div * trader.hold_volume / 10
        gain_cash = cash_div * trader.hold_volume
        origin_cost = trader.hold_cost * trader.hold_volume
        trader.hold_volume += gain_stock_div
        new_cost = origin_cost - gain_cash
        trader.hold_cost = (
            new_cost / trader.hold_volume if trader.hold_volume != 0 else 0
        )
        trader.UnrealizedProfit = round(
            (
                trader.trade_price * (1 - trader.tax - trader.fee)
                - trader.hold_cost
            )
            * trader.hold_volume,
            2,
        )
        trader.RealizedProfit += gain_cash
        trader.EverytimeProfit = trader.RealizedProfit + trader.UnrealizedProfit

    def __compute_final_stats(self):
        self._final_stats["MeanProfit"] = np.mean(
            self._results["EverytimeProfit"]
        )
        self._final_stats["MaxLoss"] = np.min(self._results["EverytimeProfit"])
        self._final_stats["FinalProfit"] = self._results[
            "EverytimeProfit"
        ].values[-1]
        self._final_stats["MeanProfitPer[%]"] = round(
            self._final_stats["MeanProfit"] / self.trader_fund * 100, 2
        )
        self._final_stats["FinalProfitPer[%]"] = round(
            self._final_stats["FinalProfit"] / self.trader_fund * 100, 2
        )
        self._final_stats["MaxLossPer[%]"] = round(
            self._final_stats["MaxLoss"] / self.trader_fund * 100, 2
        )

    def get_final_stats(self) -> pd.Series():
        return self._final_stats

    def get_results(self) -> pd.DataFrame():
        return self._results

    def plot(
        self,
        title: str = "Backtest Result",
        xlabel: str = "Time",
        ylabel: str = "Profit",
        grid: bool = True,
    ):
        try:
            import matplotlib.pyplot as plt
            import matplotlib.gridspec as gridspec
        except ImportError:
            raise ImportError("You must install matplotlib to plot importance")

        fig = plt.figure(figsize=(12, 8))
        gs = gridspec.GridSpec(4, 1, figure=fig)
        ax = fig.add_subplot(gs[:2, :])
        xpos = self._results.index
        ax.plot("UnrealizedProfit", data=self._results, marker="", alpha=0.8)
        ax.plot("RealizedProfit", data=self._results, marker="", alpha=0.8)
        ax.plot("EverytimeProfit", data=self._results, marker="", alpha=0.8)
        ax.grid(grid)
        ax.legend(loc=2)
        axx = ax.twinx()
        axx.bar(
            xpos,
            self._results["hold_volume"],
            alpha=0.2,
            label="hold_volume",
            color="pink",
        )
        axx.legend(loc=3)
        ax2 = fig.add_subplot(gs[2:, :], sharex=ax)
        ax2.plot(
            "trade_price",
            data=self._results,
            marker="",
            label="open",
            alpha=0.8,
        )
        ax2.plot(
            "hold_cost",
            data=self._results,
            marker="",
            label="hold_cost",
            alpha=0.8,
        )
        # TODO: add signal plot
        ax2.legend(loc=2)
        ax2.grid(grid)
        if title is not None:
            ax.set_title(title)
        if xlabel is not None:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        plt.show()



from FinMind.BackTestSystem.Strategies.Bias import Bias
from FinMind.BackTestSystem.Strategies.ContinueHolding import ContinueHolding
from FinMind.BackTestSystem.Strategies.InstitutionalInvestorsFollower import (
    InstitutionalInvestorsFollower,
)
from FinMind.BackTestSystem.Strategies.Kd import Kd
from FinMind.BackTestSystem.Strategies.KdCrossOver import KdCrossOver
from FinMind.BackTestSystem.Strategies.MacdCrossOver import MacdCrossOver
from FinMind.BackTestSystem.Strategies.MaCrossOver import MaCrossOver
from FinMind.BackTestSystem.Strategies.MaxMinPeriodBias import MaxMinPeriodBias
from FinMind.BackTestSystem.Strategies.NaiveKd import NaiveKd
from FinMind.BackTestSystem.Strategies.ShortSaleMarginPurchaseRatio import (
    ShortSaleMarginPurchaseRatio,
)




"""
G8
俄羅斯、美國、加拿大、英國、法國、德國、義大利及日本
"""
TABLE = "GovernmentBonds"
import os
import requests
import sys
import pandas as pd
import re
import datetime
from lxml import etree

PATH = "/".join(os.path.abspath(__file__).split("/")[:-2])
sys.path.append(PATH)
import Crawler.BasedClass as cbaseclass


"""
self = Crawler()

"""


class Crawler(cbaseclass.Crawler):
    def create_loop_list(self):
        def get_data_id_name(curl):

            headers = {
                "Accept": "*/*",
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": "zh-TW,zh;q=0.9,en-US;q=0.8,en;q=0.7",
                "Connection": "keep-alive",
                "Host": "www.investing.com",
                "Referer": curl,
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.84 Safari/537.36",
                "X-Requested-With": "XMLHttpRequest",
            }

            res = requests.get(curl, verify=True, headers=headers)
            tem_data_id = re.findall('data-id="[0-9]+"', res.text)
            tem_data_id = [di.replace("data-id=", "") for di in tem_data_id]
            page = etree.HTML(res.text)
            data_id = []
            data_name = []
            for di in tem_data_id:
                tem = page.xpath("//span[@data-id={}]".format(di))
                if len(tem) > 0:
                    data_id.append(tem[0].attrib["data-id"])
                    data_name.append(tem[0].attrib["data-name"])

            return data_id, data_name

        def get_country_url():
            index_url = "https://www.investing.com/rates-bonds/"
            headers = {
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": "zh-TW,zh;q=0.9,en-US;q=0.8,en;q=0.7",
                "Cache-Control": "max-age=0",
                "Connection": "keep-alive",
                "Host": "www.investing.com",
                "Upgrade-Insecure-Requests": "1",
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.84 Safari/537.36",
            }
            res = requests.get(index_url, verify=True, headers=headers)

            data_country_id = re.findall('data-country-id="[0-9]+"', res.text)
            data_country_id = [
                dci.replace("data-country-id=", "") for dci in data_country_id
            ]
            page = etree.HTML(res.text)
            tem = []
            for dci in data_country_id:
                tem.append(
                    page.xpath("//option[@data-country-id={}]".format(dci))[0]
                )

            url = [
                "https://www.investing.com" + te.attrib["value"] for te in tem
            ]
            # G8 and china
            select = [
                "canada",
                "china",
                "france",
                "germany",
                "japan",
                "russia",
                "uk",
                "usa",
                "italy",
            ]
            country_url = []
            for i in range(len(url)):
                tem = url[i].replace(
                    "https://www.investing.com/rates-bonds/", ""
                )
                tem = tem.replace("-government-bonds", "")
                if tem in select:
                    country_url.append(url[i])
            return country_url

        # main
        country_url = get_country_url()
        loop_list = []
        for curl in country_url:  # curl = country_url[0]
            print(curl)
            data_id, data_name = get_data_id_name(curl)
            for i in range(len(data_id)):
                loop_list.append([data_id[i], data_name[i]])

        return loop_list

    def get_st_date(self, data_name, cid):
        return "01/01/1970"

    def get_end_date(self):

        end_date = datetime.datetime.now().date()
        end_date = end_date + datetime.timedelta(-1)
        y = str(end_date.year)
        m = (
            str(end_date.month)
            if end_date.month > 9
            else "0" + str(end_date.month)
        )
        d = str(end_date.day) if end_date.day > 9 else "0" + str(end_date.day)

        return "{}/{}/{}".format(m, d, y)

    def crawler(self, loop):  # loop = ['23681', 'Germany 3 Month']
        def get_value(tem):

            date = int(tem[0].attrib["data-real-value"])
            date = int(date / 60 / 60 / 24)
            date = str(
                datetime.date(1970, 1, 1) + datetime.timedelta(days=date)
            )
            v = [float(tem[i].text) for i in range(1, 5) if tem[i].text != None]
            if len(v) == 0:
                return pd.DataFrame()
            price, Open, High, Low = v

            Change = float(tem[5].text.replace("%", "").replace(",", "")) / 100

            data = pd.DataFrame([date, price, Open, High, Low, Change]).T
            return data

        cid, data_name = loop
        header = data_name + " Bond Yield Historical Data"
        st_date, end_date = (
            self.get_st_date(data_name, cid),
            self.get_end_date(),
        )
        bonds_url = "https://www.investing.com/instruments/HistoricalDataAjax"

        form_data = {
            "curr_id": cid,
            "header": header,
            "st_date": st_date,
            "end_date": end_date,
            "interval_sec": "Daily",
            "sort_col": "date",
            "sort_ord": "DESC",
            "action": "historical_data",
        }

        headers = {
            "Accept": "text/plain, */*; q=0.01",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "zh-TW,zh;q=0.9,en-US;q=0.8,en;q=0.7",
            "Connection": "keep-alive",
            "Content-Length": "192",
            "Content-Type": "application/x-www-form-urlencoded",
            "Host": "www.investing.com",
            "Origin": "https://www.investing.com",
            "Referer": "https://www.investing.com/rates-bonds/france-1-month-bond-yield-historical-data",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.84 Safari/537.36",
            "X-Requested-With": "XMLHttpRequest",
        }
        print("requests post")
        res = requests.post(
            bonds_url, verify=True, headers=headers, data=form_data
        )
        print("data clean")
        page = etree.HTML(res.text)
        tr_path = page.xpath("//tr")

        colname = [col.text for col in tr_path[0].xpath("//th")]
        colname = [c.replace(" %", "Percent") for c in colname]
        colname = ["date" if c == "Date" else c for c in colname]
        data = pd.DataFrame()
        td_path = page.xpath("//tr//td")
        for i in range(0, len(td_path) - 6, 6):
            tem = td_path[i : i + 6]
            value = get_value(tem)
            if len(value) > 0:
                data = data.append(value)

        if len(data) > 0:
            data.columns = colname
            data["name"] = "{}".format(data_name)
            # data['data_id'] = cid
            data = data.sort_values("date")
            data.index = range(len(data))

        return data



import numpy as np
import pandas as pd
from ta.momentum import StochasticOscillator

from FinMind.BackTestSystem.BaseClass import Strategy
from FinMind.Data import Load


class ShortSaleMarginPurchaseRatio(Strategy):
    """
    url: "https://blog.above.tw/2018/08/15/%E7%B1%8C%E7%A2%BC%E9%9D%A2%E7%9A%84%E9%97%9C%E9%8D%B5%E6%8C%87%E6%A8%99%E6%9C%89%E5%93%AA%E4%BA%9B%EF%BC%9F/"
    summary:
        策略概念: 券資比越高代表散戶看空，法人買超股票會上漲，這時候賣可以跟大部分散戶進行相反的操作，反之亦然
        策略規則: 券資比>=30% 且法人買超股票, 賣
                券資比<30% 且法人賣超股票 買
    """

    ShortSaleMarginPurchaseTodayRatioThreshold = 0.3

    def init(self, base_data):
        base_data = base_data.sort_values("date")

        stock_id = base_data["stock_id"].unique()
        start_date = base_data["date"].min()
        end_date = base_data["date"].max()

        TaiwanStockMarginPurchaseShortSale = Load.FinData(
            dataset="TaiwanStockMarginPurchaseShortSale",
            select=stock_id,
            date=start_date,
            end_date=end_date,
        )

        TaiwanStockMarginPurchaseShortSale[
            ["ShortSaleTodayBalance", "MarginPurchaseTodayBalance"]
        ] = TaiwanStockMarginPurchaseShortSale[
            ["ShortSaleTodayBalance", "MarginPurchaseTodayBalance"]
        ].astype(
            int
        )
        TaiwanStockMarginPurchaseShortSale[
            "ShortSaleMarginPurchaseTodayRatio"
        ] = (
            TaiwanStockMarginPurchaseShortSale["ShortSaleTodayBalance"]
            / TaiwanStockMarginPurchaseShortSale["MarginPurchaseTodayBalance"]
        )

        InstitutionalInvestorsBuySell = Load.FinData(
            dataset="InstitutionalInvestorsBuySell",
            select=stock_id,
            date=start_date,
            end_date=end_date,
        )

        InstitutionalInvestorsBuySell[["sell", "buy"]] = (
            InstitutionalInvestorsBuySell[["sell", "buy"]].fillna(0).astype(int)
        )
        InstitutionalInvestorsBuySell = InstitutionalInvestorsBuySell.groupby(
            ["date", "stock_id"], as_index=False
        ).agg({"buy": np.sum, "sell": np.sum})
        InstitutionalInvestorsBuySell["diff"] = (
            InstitutionalInvestorsBuySell["buy"]
            - InstitutionalInvestorsBuySell["sell"]
        )
        base_data = pd.merge(
            base_data,
            InstitutionalInvestorsBuySell[["stock_id", "date", "diff"]],
            on=["stock_id", "date"],
            how="left",
        ).fillna(0)
        base_data = pd.merge(
            base_data,
            TaiwanStockMarginPurchaseShortSale[
                ["stock_id", "date", "ShortSaleMarginPurchaseTodayRatio"]
            ],
            on=["stock_id", "date"],
            how="left",
        ).fillna(0)

        base_data.index = range(len(base_data))

        base_data["signal"] = 0
        sell_mask = (
            base_data["ShortSaleMarginPurchaseTodayRatio"]
            >= self.ShortSaleMarginPurchaseTodayRatioThreshold
        ) & (base_data["diff"] > 0)
        base_data.loc[sell_mask, "signal"] = -1
        buy_mask = (
            base_data["ShortSaleMarginPurchaseTodayRatio"]
            < self.ShortSaleMarginPurchaseTodayRatioThreshold
        ) & (base_data["diff"] < 0)
        base_data.loc[buy_mask, "signal"] = 1
        return base_data



import typing

import numpy as np
import pandas as pd
import requests

from FinMind.config import API_HOST


def FinData(
    dataset: str,
    select: str = "",
    date: str = "",
    end_date: str = "",
    user_id: str = "",
    password: str = "",
    url: str = f"{API_HOST}/data",
) -> pd.DataFrame:

    params = {
        "dataset": dataset,
        "stock_id": select,
        "date": date,
        "end_date": end_date,
        "user_id": user_id,
        "password": password,
    }
    res = requests.get(url, verify=True, params=params)

    data = pd.DataFrame()
    try:
        data = res.json()
        if data["status"] == 200:
            data = pd.DataFrame(data["data"])
    except Exception:
        pass

    return data


def FinDataList(
    dataset: str,
    user_id: str = "",
    password: str = "",
    list_url=f"{API_HOST}/datalist",
) -> typing.List:

    params = {
        "dataset": dataset,
        "user_id": user_id,
        "password": password,
    }
    res = requests.get(list_url, verify=True, params=params)

    data = res.json()
    if data["status"] == 200:
        data = data["data"]

    return data


def CrawlerStockInfo(dataset: str = "") -> pd.DataFrame:
    data = FinData(dataset)
    return data


def transpose(data, var="type"):
    select_variable = "stock_id"
    date_list = list(np.unique(data["date"]))
    data1 = pd.DataFrame()
    select_var_list = list(np.unique(data[select_variable]))
    for date in date_list:
        for select_var in select_var_list:
            data2 = data.loc[
                (data["date"] == date) & (data[select_variable] == select_var),
                [var, "value"],
            ]
            data2.index = data2[var]
            del data2[var]
            data2 = data2.T
            data2.index = range(len(data2))
            data2.columns = list(data2.columns)
            data2["stock_id"] = select_var
            data2["date"] = date
            data1 = data1.append(data2)

    data1.index = range(len(data1))
    return data1


def translation(
    dataset: str, user_id: str = "", password: str = ""
) -> pd.DataFrame:
    url = f"{API_HOST}/translation"
    params = {
        "dataset": dataset,
        "user_id": user_id,
        "password": password,
    }
    res = requests.get(url, verify=True, params=params)
    data = res.json()
    if data["status"] == 200:
        data = pd.DataFrame(data["data"])
    return data



from ta.trend import MACD

from FinMind.BackTestSystem.BaseClass import Strategy


class MacdCrossOver(Strategy):
    """
    url:https://www.cmoney.tw/learn/course/technicals/topic/750

    summary:
    MACD 指標由 DIF 與 MACD 兩條線組成
    DIF（快）短期，判斷股價趨勢的變化
    MACD（慢）長期，判斷股價大趨勢
    短期線 突破 長期線(黃金交叉)，進場
    長期線 突破 短期線(死亡交叉)，出場
    """

    def init(self, base_data):
        base_data = base_data.sort_values("date")

        macd = MACD(close=base_data["close"], n_slow=26, n_fast=12, n_sign=9)

        base_data["macd_diff"] = macd.macd_diff()
        base_data["macd_signal"] = macd.macd_signal()

        base_data["fs_dif"] = base_data["macd_diff"] - base_data["macd_signal"]
        base_data["bool_signal"] = base_data["fs_dif"].map(
            lambda x: 1 if x > 0 else -1
        )
        base_data["bool_signal_shift1"] = base_data["bool_signal"].shift(1)

        base_data["signal"] = 0
        base_data.loc[
            (
                (base_data["bool_signal"] > 0)
                & (base_data["bool_signal_shift1"] < 0)
            ),
            "signal",
        ] = 1
        base_data.loc[
            (
                (base_data["bool_signal"] < 0)
                & (base_data["bool_signal_shift1"] > 0)
            ),
            "signal",
        ] = -1

        base_data.index = range(len(base_data))
        return base_data



import numpy as np
from ta.momentum import StochasticOscillator

from FinMind.BackTestSystem.BaseClass import Strategy


class Kd(Strategy):
    """
    url: "https://www.mirrormedia.mg/story/20180719fin012/"
    summary:
        日KD 80 20
        日K線 <= 20 進場
        日K線 >= 80 出場
    """

    kdays = 9
    kd_upper = 80
    kd_lower = 20

    def init(self, base_data):
        base_data = base_data.sort_values("date")

        kd = StochasticOscillator(
            high=base_data["max"],
            low=base_data["min"],
            close=base_data["close"],
            n=self.kdays,
        )
        rsv_ = kd.stoch().fillna(50)

        _k = np.zeros(base_data.shape[0])
        _d = np.zeros(base_data.shape[0])
        for i, r in enumerate(rsv_):
            if i == 0:
                _k[i] = 50
                _d[i] = 50
            else:
                _k[i] = _k[i - 1] * 2 / 3 + r / 3
                _d[i] = _d[i - 1] * 2 / 3 + _k[i] / 3

        base_data["K"] = _k
        base_data["D"] = _d

        base_data.index = range(len(base_data))

        base_data["signal"] = 0
        base_data.loc[base_data["K"] <= self.kd_lower, "signal"] = 1
        base_data.loc[base_data["K"] >= self.kd_upper, "signal"] = -1
        return base_data



import datetime
import numpy as np
import requests
import time


def requests_get(url, header):
    try:
        res = requests.get(url, verify=False, timeout=10, headers=header)
        return res
    except Exception as e:
        if "Max retries exceeded" in str(e) or "Read timed out" in str(e):
            time.sleep(60)


def requests_post(url, header, form_data):
    try:
        res = requests.post(
            url, verify=False, timeout=10, headers=header, data=form_data
        )
        return res
    except Exception as e:
        if "Max retries exceeded" in str(e) or "Read timed out" in str(e):
            time.sleep(60)
            raise


def get_time():
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))


class Crawler:
    def date2days(self, date):
        # date = '2018-08-03'
        date = datetime.datetime.strptime(date, "%Y-%m-%d").date()
        value = (date - datetime.date(1970, 1, 1)).days
        value = value * 60 * 60 * 24 * 1000
        return value

    def days2date(self, day):
        # day = 631497600000
        # 60s = 1min
        # 60min = 1hr
        day = int(day)
        day = int(day / 1000 / 60 / 60 / 24)
        value = datetime.date(1970, 1, 1) + datetime.timedelta(days=day)
        return value

    def millisecond2date(self, ms):
        # ms = 1559489350000
        ms = int(ms)
        date = str(self.days2date(ms))
        days = int(ms / 1000 / 60 / 60 / 24)

        second = ms / 1000 - days * 60 * 60 * 24
        hour = int(second / 60 / 60)
        minute = int((second - hour * 60 * 60) / 60) - 1

        if hour < 0:
            hour = "00"
        elif hour < 10:
            hour = "0" + str(hour)
        else:
            hour = str(hour)

        if minute < 0:
            minute = "00"
        elif minute < 10:
            minute = "0" + str(minute)
        else:
            minute = str(minute)
        # second = (second - hour*60*60 - minute*60)#hour
        value = date + " " + hour + ":" + minute + ":00"
        return value

    def millisecond2date2(self, ms):
        # ms = 1566898740000
        ms = int(ms)
        date = str(self.days2date(ms))
        days = int(ms / 1000 / 60 / 60 / 24)

        second = ms / 1000 - days * 60 * 60 * 24
        hour = int(second / 60 / 60)
        minute = int((second - hour * 60 * 60) / 60)
        second = int(second - hour * 60 * 60 - minute * 60)

        if hour < 0:
            hour = "00"
        elif hour < 10:
            hour = "0" + str(hour)
        else:
            hour = str(hour)

        if minute < 0:
            minute = "00"
        elif minute < 10:
            minute = "0" + str(minute)
        else:
            minute = str(minute)
        # second = (second - hour*60*60 - minute*60)#hour
        if second < 10:
            second = "0{}".format(second)
        value = "{} {}:{}:{}".format(date, hour, minute, second)
        return value

    def date2millisecond(self, date):
        # date = '2019-06-02 15:30:00'
        date = datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S")
        date = date + datetime.timedelta(minutes=1)
        second = date - datetime.datetime(1970, 1, 1, 0, 0, 0)

        second = second.days * 24 * 60 * 60 + second.seconds
        # ms = ms*1000

        return second

    def create_date(self, start, today=False):  # start = '2018-07-31'
        start = datetime.datetime.strptime(
            start, "%Y-%m-%d"
        ).date() + datetime.timedelta(days=1)
        end = datetime.date.today()

        day_len = (end - start).days
        if today:
            day_len = (end - start).days + 1
        date = [
            str(start + datetime.timedelta(days=dat)) for dat in range(day_len)
        ]
        return date

    def remove_outlier(self, data, var_name):

        value = list(data[var_name])
        mean = np.mean(value, axis=0)
        sd = np.std(value, axis=0)
        if sd < 1:
            return data

        _bool = []
        for x in value:
            if (5 * mean) > x > (-5 * mean):
                _bool.append(True)
            else:
                _bool.append(False)

        data = data[_bool]
        return data


def change_chinese_date_us(d):
    y, m, d = [int(x) for x in d.split("/")]
    y = y + 1911
    date = datetime.date(y, m, d)
    return date



from setuptools import setup, find_packages
from io import open
import os

_version = os.environ.get("CI_COMMIT_TAG", "1.2.0.dev6")
here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


requirements = [
    "ta", "requests", "importlib_metadata", "matplotlib", "pandas"
]

setup(
    name="FinMind",  # Required
    version=_version,  # Required
    description="financial mining",  # Optional
    long_description=long_description,  # Optional
    long_description_content_type="text/markdown",  # Optional (see note above)
    url="https://github.com/linsamtw",  # Optional
    author="linsam",  # Optional
    author_email="samlin266118@gmail.com",  # Optional
    classifiers=[  # Optional
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
    ],
    keywords="financial, python",  # Optional
    packages=find_packages(exclude=["importlib", "ta"]),
    install_requires=requirements,
    project_urls={  # Optional
        "documentation": "https://linsamtw.github.io/FinMindDoc/",
        "Source": "https://github.com/linsamtw/FinMind",
    },
)



TABLE = "RawMaterialFuturesPrices"

import os
import requests
import sys
import pandas as pd
import datetime
import re
from lxml import etree

PATH = "/".join(os.path.abspath(__file__).split("/")[:-2])
sys.path.append(PATH)
import Crawler.BasedClass as cbaseclass

"""
self = Crawler()
"""


class Crawler(cbaseclass.Crawler):
    def __init__(self):
        super(Crawler, self).__init__()

    def create_loop_list(self):
        # self.based_url = 'https://www.investing.com/commodities/'
        kind_list = ["meats", "grains", "energies", "softs", "metals"]
        loop_list = []
        for kind in kind_list:
            print(kind)
            index_url = "https://www.investing.com/commodities/" + kind
            headers = {
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": "zh-TW,zh;q=0.9,en-US;q=0.8,en;q=0.7",
                "Cache-Control": "max-age=0",
                "Connection": "keep-alive",
                "Host": "www.investing.com",
                "Upgrade-Insecure-Requests": "1",
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.84 Safari/537.36",
            }
            res = requests.get(index_url, verify=True, headers=headers)

            tem = re.findall(
                ' data-name="[A-Za-z ]+" data-id="[0-9]+" ', res.text
            )
            futures_id_list = [re.findall("[0-9]+", te)[0] for te in tem]
            data_name_list = [
                re.findall('"[A-Za-z ]+"', te)[0].replace('"', "") for te in tem
            ]
            [
                loop_list.append([futures_id_list[i], data_name_list[i]])
                for i in range(len(futures_id_list))
            ]
        return loop_list

    def get_st_date(self, futures_id, data_name):
        return "01/01/1970"
        # return '01/01/2018'

    def get_end_date(self):
        end_date = datetime.datetime.now().date()
        end_date = end_date + datetime.timedelta(-1)
        y = str(end_date.year)
        m = (
            str(end_date.month)
            if end_date.month > 9
            else "0" + str(end_date.month)
        )
        d = str(end_date.day) if end_date.day > 9 else "0" + str(end_date.day)

        return m + "/" + d + "/" + y

    def crawler(self, loop):  # loop = ['49769', 'Brent Oil Futures']
        def get_value(tem):

            date = int(tem[0].attrib["data-real-value"])
            date = int(date / 60 / 60 / 24)
            date = str(
                datetime.date(1970, 1, 1) + datetime.timedelta(days=date)
            )
            v = [
                float(tem[i].attrib["data-real-value"].replace(",", ""))
                for i in range(1, 6)
            ]
            price, Open, High, Low, Vol = v

            Change = float(tem[6].text.replace("%", "").replace(",", "")) / 100

            data = pd.DataFrame([date, price, Open, High, Low, Vol, Change]).T

            return data

        # -------------------------------------------------------------------
        futures_id, data_name = loop
        header = data_name + " Historical Data"
        st_date, end_date = (
            self.get_st_date(futures_id, data_name),
            self.get_end_date(),
        )

        bonds_url = "https://www.investing.com/instruments/HistoricalDataAjax"
        form_data = {
            "curr_id": futures_id,
            "header": header,
            "st_date": st_date,
            "end_date": end_date,
            "interval_sec": "Daily",
            "sort_col": "date",
            "sort_ord": "DESC",
            "action": "historical_data",
        }

        headers = {
            "Accept": "text/plain, */*; q=0.01",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "zh-TW,zh;q=0.9,en-US;q=0.8,en;q=0.7",
            "Connection": "keep-alive",
            "Content-Length": "183",
            "Content-Type": "application/x-www-form-urlencoded",
            "Host": "www.investing.com",
            "Origin": "https://www.investing.com",
            "Referer": "https://www.investing.com/commodities/brent-oil-historical-data",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.84 Safari/537.36",
            "X-Requested-With": "XMLHttpRequest",
        }
        print("requests post")
        res = requests.post(
            bonds_url, verify=True, headers=headers, data=form_data
        )
        print("data clean")
        page = etree.HTML(res.text)
        colname = page.xpath("//tr//th")

        colname = [
            c.text.replace(" %", "Percent").replace(".", "") for c in colname
        ]
        colname = ["date" if c == "Date" else c for c in colname]

        data = pd.DataFrame()
        td_path = page.xpath("//tr//td")
        for i in range(0, len(td_path) - 7, 7):
            tem = td_path[i : i + 7]
            value = get_value(tem)
            data = data.append(value)

        if len(data) > 0:
            data.columns = colname
            data["name"] = data_name
            # data['data_id'] = futures_id
            data = data.sort_values("date")
            data.index = range(len(data))

        return data



from ta.trend import SMAIndicator

from FinMind.BackTestSystem.BaseClass import Strategy


class MaxMinPeriodBias(Strategy):
    """
    url: http://www.bituzi.com/2013/03/bias.html

    summary:
    乖離率加上最近 k 天最大最小值進出法
    單存乖離率來判斷進出相對而言比較不穩定，因此多一個限制是跟最近 k 天最大最小值股價做比較來段進出
    負乖離表示股價 低 於過去一段時間平均價且股價大於過去k天最大值，意味著股價相對過去 低 且即將走高 ，則選擇進場
    正乖離表示股價 高 於過去一段時間平均價且股價大於過去k天最小值，意味著股價相對過去 高 且即將走低，則選擇出場
    相對於單存乖離率而言來的保守
    """

    ma_days = 24
    last_k_days = 5
    bais_lower = -7
    bais_upper = 8

    def init(self, base_data):
        base_data = base_data.sort_values("date")

        base_data[f"ma{self.ma_days}"] = SMAIndicator(
            base_data["close"], self.ma_days
        ).sma_indicator()

        base_data["bias"] = (
            (base_data["close"] - base_data[f"ma{self.ma_days}"])
            / base_data[f"ma{self.ma_days}"]
        ) * 100

        base_data[f"max_last_k_days{self.last_k_days}"] = (
            base_data["close"].shift(1).rolling(window=self.last_k_days).max()
        )

        base_data[f"min_last_k_days{self.last_k_days}"] = (
            base_data["close"].shift(1).rolling(window=self.last_k_days).min()
        )

        base_data["signal"] = 0
        base_data.loc[
            (
                (base_data["bias"] < self.bais_lower)
                & (
                    base_data["close"]
                    > base_data[f"max_last_k_days{self.last_k_days}"]
                )
            ),
            "signal",
        ] = 1
        base_data.loc[
            (
                (base_data["bias"] > self.bais_upper)
                & (
                    base_data["close"]
                    < base_data[f"min_last_k_days{self.last_k_days}"]
                )
            ),
            "signal",
        ] = -1

        return base_data



import numpy as np
import pandas as pd
from ta.momentum import StochasticOscillator

from FinMind.BackTestSystem.BaseClass import Strategy
from FinMind.Data import Load


class InstitutionalInvestorsFollower(Strategy):
    """
    url: "https://www.finlab.tw/%E8%85%A6%E5%8A%9B%E6%BF%80%E7%9B%AA%E7%9A%84%E5%A4%96%E8%B3%87%E7%AD%96%E7%95%A5%EF%BC%81/"
    summary:
        策略概念: 法人大量買超會導致股價上漲, 賣超反之
        策略規則: 三大法人大量買超隔天就賣，大量賣超就買
    """

    def init(self, base_data):
        base_data = base_data.sort_values("date")
        base_data.index = range(len(base_data))

        stock_id = base_data["stock_id"].unique()
        start_date = base_data["date"].min()
        end_date = base_data["date"].max()

        InstitutionalInvestorsBuySell = Load.FinData(
            dataset="InstitutionalInvestorsBuySell",
            select=stock_id,
            date=start_date,
            end_date=end_date,
        )
        InstitutionalInvestorsBuySell = InstitutionalInvestorsBuySell.groupby(
            ["date", "stock_id"], as_index=False
        ).agg({"buy": np.sum, "sell": np.sum})
        InstitutionalInvestorsBuySell["diff"] = (
            InstitutionalInvestorsBuySell["buy"]
            - InstitutionalInvestorsBuySell["sell"]
        )
        base_data = pd.merge(
            base_data,
            InstitutionalInvestorsBuySell[["stock_id", "date", "diff"]],
            on=["stock_id", "date"],
            how="left",
        ).fillna(0)

        base_data["signal_info"] = self.detect_Abnormal_Peak(
            y=base_data["diff"].values, lag=10, threshold=3, influence=0.35,
        )["signals"]

        base_data["signal"] = 0
        base_data.loc[base_data["signal_info"] == -1, "signal"] = 1
        base_data.loc[base_data["signal_info"] == 1, "signal"] = -1
        return base_data

    def detect_Abnormal_Peak(
        self, y: np.array, lag: int, threshold: float, influence: float
    ):
        signals = np.zeros(len(y))
        filteredY = np.array(y)
        avgFilter = [0] * len(y)
        stdFilter = [0] * len(y)
        avgFilter[lag - 1] = np.mean(y[0:lag])
        stdFilter[lag - 1] = np.std(y[0:lag])
        for i in range(lag, len(y)):
            if abs(y[i] - avgFilter[i - 1]) > threshold * stdFilter[i - 1]:
                if y[i] > avgFilter[i - 1]:
                    signals[i] = 1
                else:
                    signals[i] = -1

                filteredY[i] = (
                    influence * y[i] + (1 - influence) * filteredY[i - 1]
                )
                avgFilter[i] = np.mean(filteredY[(i - lag + 1) : i + 1])
                stdFilter[i] = np.std(filteredY[(i - lag + 1) : i + 1])
            else:
                signals[i] = 0
                filteredY[i] = y[i]
                avgFilter[i] = np.mean(filteredY[(i - lag + 1) : i + 1])
                stdFilter[i] = np.std(filteredY[(i - lag + 1) : i + 1])

        return dict(
            signals=np.asarray(signals),
            avgFilter=np.asarray(avgFilter),
            stdFilter=np.asarray(stdFilter),
        )



from ta.trend import SMAIndicator

from FinMind.BackTestSystem.BaseClass import Strategy


class MaCrossOver(Strategy):
    """
    url:
    "https://www.cmoney.tw/learn/course/technicalanalysisfast/topic/1811"

    summary:
    均線黃金交叉
    以短線操作來說，當 5日均線 向上突破 20日均線
    也就是短期的平均買進成本大於長期平均成本
    代表短期買方的力道較大，市場上大多數人獲利
    市場易走出「多頭」的趨勢，進而帶動長期均線向上，讓股價上漲機率較大
    短期線 突破 長期線(黃金交叉)，進場
    長期線 突破 短期線(死亡交叉)，出場
    """

    ma_fast_days = 10
    ma_slow_days = 30

    def init(self, base_data):
        base_data = base_data.sort_values("date")

        base_data[f"ma{self.ma_fast_days}"] = SMAIndicator(
            base_data["close"], self.ma_fast_days
        ).sma_indicator()

        base_data[f"ma{self.ma_slow_days}"] = SMAIndicator(
            base_data["close"], self.ma_slow_days
        ).sma_indicator()

        base_data["madiff"] = (
            base_data[f"ma{self.ma_fast_days}"]
            - base_data[f"ma{self.ma_slow_days}"]
        )
        base_data["bool_signal"] = base_data["madiff"].map(
            lambda x: 1 if x > 0 else -1
        )
        base_data["bool_signal_shift1"] = (
            base_data["bool_signal"].shift(1).fillna(0)
        )

        base_data["bool_signal_shift1"] = base_data[
            "bool_signal_shift1"
        ].astype(int)

        base_data["signal"] = 0
        base_data.loc[
            (
                (base_data["bool_signal"] > 0)
                & (base_data["bool_signal_shift1"] < 0)
            ),
            "signal",
        ] = 1
        base_data.loc[
            (
                (base_data["bool_signal"] < 0)
                & (base_data["bool_signal_shift1"] > 0)
            ),
            "signal",
        ] = -1
        return base_data



import numpy as np
from ta.momentum import StochasticOscillator

from FinMind.BackTestSystem.BaseClass import Strategy


class KdCrossOver(Strategy):
    """
    url: "http://smart.businessweekly.com.tw/Reading/WebArticle.aspx?id=68129&p=2"
    summary: 日KD黃金交叉和死亡交叉
            日K線 小於 日D線 翻轉成 日K線 大於 日D線 稱為黃金交叉
            日K線 大於 日D線 翻轉成 日K線 小於 日D線 稱為死亡交叉
            黃金交叉進場，死亡交叉出場
    """

    kdays = 9

    def init(self, base_data):
        base_data = base_data.sort_values("date")

        kd = StochasticOscillator(
            high=base_data["max"],
            low=base_data["min"],
            close=base_data["close"],
            n=self.kdays,
        )
        rsv_ = kd.stoch().fillna(50)

        _k = np.zeros(base_data.shape[0])
        _d = np.zeros(base_data.shape[0])
        for i, r in enumerate(rsv_):
            if i == 0:
                _k[i] = 50
                _d[i] = 50
            else:
                _k[i] = _k[i - 1] * 2 / 3 + r / 3
                _d[i] = _d[i - 1] * 2 / 3 + _k[i] / 3

        base_data["K"] = _k
        base_data["D"] = _d

        base_data.index = range(len(base_data))

        base_data["diff"] = base_data["K"] - base_data["D"]
        base_data.loc[(base_data.index < self.kdays), "diff"] = np.nan

        base_data["bool_diff"] = base_data["diff"].map(
            lambda x: 1 if x >= 0 else (-1 if x < 0 else 0)
        )
        base_data["bool_diff_shift1"] = (
            base_data["bool_diff"].shift(1).fillna(0).astype(int)
        )

        base_data["signal"] = 0
        base_data.loc[
            (
                (base_data["bool_diff"] > 0)
                & (base_data["bool_diff_shift1"] < 0)
            ),
            "signal",
        ] = 1
        base_data.loc[
            (
                (base_data["bool_diff"] < 0)
                & (base_data["bool_diff_shift1"] > 0)
            ),
            "signal",
        ] = -1
        return base_data
